#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <getopt.h>
#include <inttypes.h>

#include <intrepidcs.h>

#define ERR_LEN 256
#define PSTR_LEN 100

static void usage(void);

int main(int argc, char *argv[])
{
	int packet_valid;
	unsigned char b1, b2, b3;
	char err[ERR_LEN], pstr[PSTR_LEN], opt, write_flag = 0;
	char *nextCNFbyte;
	const char *port;
	struct ics_interface *iface;
	struct ics_packet p = {ICS_NETID_DEVICE, 0xB, {0xA, 0, 0x08, 0, 0, 0, 0, 0, 0, 0, 0}}, disable_packet;

	/*
	 * Formatting for packet to set bitrate:
	 * Byte 0: Length/NETID_DEVICE (ICS_NETID_DEVICE in libintrepidcs)
	 * Byte 1: The number of bytes to follow (length - 1)
	 * Byte 2: 0 = read, 1 = write
	 * Byte 3: The adddress to read to or write from
	 * Bytes 4-12: The 8 bytes to write in the case of a write
	 * 
	 * For ics_packet, this translates to an initializer of:
	 * {ICS_NETID_DEVICE, length of following array, {count of bytes 
	 * to follow, 0 read / 1 write, address, byte 1, byte 2, byte 3,
	 * byte 4, byte 5, byte 6, byte 7, byte 8}};
	 */
	if(argc < 2)
	{
		usage();
		return EXIT_FAILURE;
	}

	 while( (opt = getopt(argc, argv, "b:p:")) != -1)
	 {
	 	switch(opt)
		{

		case 'b': //bytes to write; if absent, just read CNF registers
			b1 = (char)strtoul(optarg, &nextCNFbyte, 16);
			b2 = (char)strtoul(nextCNFbyte + 1, &nextCNFbyte, 16);
			b3 = (char)strtoul(nextCNFbyte + 1, &nextCNFbyte, 16);
			write_flag = 1;
			break;
		case 'p':
			port = optarg;
			break;
		default:
			usage();
			return EXIT_FAILURE;
		}
	}

	iface = ics_interface_init(port, err, ERR_LEN);
	if(iface == NULL)
	{
		printf("%s\n", err);
		return EXIT_FAILURE;
	}

	/* We need to disable network communications before changing settings */
	ics_packet_create_can_disable(&disable_packet);
	if(ics_interface_write_packet(iface, &disable_packet) == 0)
	{
		printf("Couldn't write disable packet: %s\n", ics_interface_err_str(iface));
		ics_interface_release(iface);
		return EXIT_FAILURE;
	};

	/* Need to throw out the response to the disable packet (0x0A) so it doesn't show up
	 * later, as well as any device status packets that crept in there */
	do { 
		if(ics_interface_read_packet(iface, &disable_packet, &packet_valid) == 0)
		{
			printf("Couldn't read packet: %s\n", ics_interface_err_str(iface));
			ics_interface_release(iface);
			return EXIT_FAILURE;
		}
		ics_packet_format(pstr, PSTR_LEN, &disable_packet);
		printf("Throwing out packet: %s\n", pstr);
	}while (disable_packet.data[0] != 0x0A);
	
	ics_packet_format(pstr, PSTR_LEN, &p);
	printf("Sending packet: %s\n", pstr);

	/* Write the packet to get data from the interface */
	if(ics_interface_write_packet(iface, &p) == 0)
	{
		printf("Couldn't write packet: %s\n", ics_interface_err_str(iface));
		ics_interface_release(iface);
		return EXIT_FAILURE;
	}

	/* Interface's response*/
	if(ics_interface_read_packet(iface, &p, &packet_valid) == 0)
	{
		printf("Couldn't read packet: %s\n", ics_interface_err_str(iface));
		ics_interface_release(iface);
		return EXIT_FAILURE;
	}
	
	ics_packet_format(pstr, PSTR_LEN, &p);
	printf("Response: %s\n", pstr);

	if(write_flag)
	{
		/* next step: set bitrate */
		p.data[0] = 0xA; /* was a 7, which means that p was a response to our 
				read request, now the number of bytes to follow
				(should always be A)*/
		p.data[1] = 0x01; /* was a 0 for read, but now we want to write */
		p.data[2] = 0x08; /* 0x08 should be left in p anyway since that's the
				address we read */

		/* note: the following registers are in reverse numerical order! (CNF3 to CNF1) */

		p.data[3] = b3; /* CNF3 as given by neoVI Explorer */
		p.data[4] = b2; /* CNF2 */
		p.data[5] = b1; /* CNF1 */
	
		/* that does it for changing the bitrate; leave everything else */
	

		if(ics_interface_write_packet(iface, &p) == 0)
		{
			printf("Couldn't change bitrate: %s\n", ics_interface_err_str(iface));
			ics_interface_release(iface);
			return EXIT_FAILURE;
		}

		ics_packet_format(pstr, PSTR_LEN, &p);
		printf("Sending packet: %s\n", pstr);

		/* Interface's response */
		if(ics_interface_read_packet(iface, &p, &packet_valid) == 0)
		{
			printf("Couldn't read packet: %s\n", ics_interface_err_str(iface));
			ics_interface_release(iface);
			return EXIT_FAILURE;
		}

		ics_packet_format(pstr, PSTR_LEN, &p);
		printf("Response: %s\n", pstr);
	}

	/* We need to reenable network communications */
	ics_packet_create_can_enable(&disable_packet);
	if(ics_interface_write_packet(iface, &disable_packet) == 0)
	{
		printf("Couldn't write enable packet: %s\n", ics_interface_err_str(iface));
		ics_interface_release(iface);
		return EXIT_FAILURE;
	};

	/* Need to get the response to the enable packet so it doesn't show up later*/
	if(ics_interface_read_packet(iface, &disable_packet, &packet_valid) == 0)
	{
		printf("Couldn't read packet: %s\n", ics_interface_err_str(iface));
		ics_interface_release(iface);
		return EXIT_FAILURE;
	}
	ics_interface_release(iface);
	return EXIT_SUCCESS;
}

static void usage(void)
{
	fprintf(stderr, "Usage: can_bitrate -b \"CNF1:CNF2:CNF3\" -p port\n");
}
